use std::collections::HashMap;
use std::sync::mpsc::Sender;
use std::time::Duration;

use color_eyre::eyre::Context as _;
use color_eyre::Result;
use reqwest::{header, Client};
use tokio::runtime;
use tokio::sync::mpsc::UnboundedReceiver;
use tracing::Instrument as _;

use crate::types::{ApiEvent, ApiTask, Event, Method};
use crate::APP_USER_AGENT;

fn header_map_to_hashmap(headers: &header::HeaderMap) -> HashMap<String, Vec<u8>> {
    let mut map = HashMap::with_capacity(headers.len());

    for (k, v) in headers {
        map.insert(k.to_string(), v.as_bytes().to_vec());
    }

    map
}

#[tracing::instrument]
async fn perform_request(client: &Client, task: &ApiTask) -> Result<Event> {
    let req = match task.method {
        Method::Get => client.get(&task.url),
        Method::Post => {
            let body = serde_json::to_vec(&task.body)
                .expect("Type `serde_json::Value` should always be serializable");
            client.post(&task.url).body(body)
        }
    };

    let res = req
        .query(&task.query)
        .headers({
            let mut headers = header::HeaderMap::new();

            for (k, v) in &task.headers {
                // Non-ASCII characters aren't supported anyways for header name, so no need to map
                // those.
                let k = header::HeaderName::from_lowercase(k.to_ascii_lowercase().as_bytes())
                    .wrap_err_with(|| format!("Invalid header name '{}'", k))?;
                let v = header::HeaderValue::from_str(v)
                    .wrap_err_with(|| format!("Invalid header value '{}'", v))?;

                headers.insert(k, v);
            }

            headers
        })
        .send()
        .await?;
    let status = res.status();
    let headers = header_map_to_hashmap(res.headers());

    let body: serde_json::Value = if let Ok(text) = res.text().await {
        serde_json::from_str(&text).unwrap_or(serde_json::Value::String(text))
    } else {
        serde_json::Value::Null
    };

    let event = ApiEvent {
        id: task.id.clone(),
        headers,
        body,
        status: status.as_u16(),
    };
    Ok(Event::Api(event))
}

#[tracing::instrument(skip_all)]
pub async fn run(mut api_rx: UnboundedReceiver<ApiTask>, event_tx: Sender<Event>) -> Result<()> {
    let (task_tx, mut task_rx) = tokio::sync::mpsc::channel(64);

    let client = Client::builder()
        .user_agent(APP_USER_AGENT)
        .build()
        .wrap_err("Failed to build HTTP client")?;

    tokio::spawn({
        let span = tracing::info_span!("api receiver");
        async move {
            while let Some(task) = api_rx.recv().await {
                tracing::trace!(id = task.id, "Received new API task");

                let _ = task_tx
                    .send(async move {
                        tokio::time::sleep(Duration::from_secs(task.delay)).await;
                        task
                    })
                    .await;
            }
            tracing::error!("API task channel closed");
        }
        .instrument(span)
    });

    while let Some(task) = task_rx.recv().await {
        let task = task.await;
        let event = perform_request(&client, &task)
            .await
            .wrap_err("Failed to perform API request")?;
        event_tx.send(event).wrap_err("Failed to send event")?;
    }

    Ok(())
}

#[tracing::instrument("api::worker", skip_all)]
pub fn worker(api_rx: UnboundedReceiver<ApiTask>, event_tx: Sender<Event>) -> Result<()> {
    let rt = runtime::Builder::new_current_thread()
        .thread_name("api")
        .enable_io()
        .enable_time()
        .build()?;

    rt.block_on(run(api_rx, event_tx))
}
